# Copyright 2022 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jwt

# From https://github.com/sigstore/fulcio/blob/b2186c0/pkg/config/config.go#L182-L201
OIDC_ISSUERS = {
    "https://accounts.google.com": "email",
    "https://oauth2.sigstore.dev/auth": "email",
    "https://token.actions.githubusercontent.com": "sub",
}
AUDIENCE = "sigstore"


class IdentityError(Exception):
    pass


class Identity:
    def __init__(self, identity_token: str) -> None:
        identity_jwt = jwt.decode(identity_token, options={"verify_signature": False})

        if "iss" not in identity_jwt:
            raise IdentityError("Identity token  missing the required 'iss' claim")

        iss = identity_jwt.get("iss")

        if iss not in OIDC_ISSUERS:
            raise IdentityError(f"Not a valid OIDC issuer: {iss!r}")

        if "aud" not in identity_jwt:
            raise IdentityError("Identity token missing the required 'aud' claim")

        aud = identity_jwt.get("aud")

        if aud != AUDIENCE:
            raise IdentityError(f"Audience should be {AUDIENCE!r}, not {aud!r}")

        proof_claim = OIDC_ISSUERS[iss]
        if proof_claim not in identity_jwt:
            raise IdentityError(
                f"Identity token missing the required {proof_claim!r} claim"
            )

        self.proof: str = str(identity_jwt.get(proof_claim))



# Copyright 2022 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from importlib import resources


def test_store_reads_fulcio_root_cert():
    fulcio_crt = resources.read_text("sigstore._store", "fulcio.crt.pem")

    assert fulcio_crt.startswith("-----BEGIN CERTIFICATE-----")
    assert fulcio_crt.endswith("-----END CERTIFICATE-----")


def test_store_reads_ctfe_pub():
    ctfe_pub = resources.read_text("sigstore._store", "ctfe.pub")

    assert ctfe_pub.startswith("-----BEGIN PUBLIC KEY-----")
    assert ctfe_pub.endswith("-----END PUBLIC KEY-----")



# Copyright 2022 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pretend
import pytest
from requests import HTTPError

from sigstore._internal.oidc import ambient


def test_detect_credential_none(monkeypatch):
    detect_github = pretend.call_recorder(lambda: None)
    monkeypatch.setattr(ambient, "detect_github", detect_github)
    assert ambient.detect_credential() is None


def test_detect_credential(monkeypatch):
    detect_github = pretend.call_recorder(lambda: "fakejwt")
    monkeypatch.setattr(ambient, "detect_github", detect_github)

    assert ambient.detect_credential() == "fakejwt"


def test_detect_github_bad_env(monkeypatch):
    # We might actually be running in a CI, so explicitly remove this.
    monkeypatch.delenv("GITHUB_ACTIONS", raising=False)

    logger = pretend.stub(debug=pretend.call_recorder(lambda s: None))
    monkeypatch.setattr(ambient, "logger", logger)

    assert ambient.detect_github() is None
    assert logger.debug.calls == [
        pretend.call("GitHub: looking for OIDC credentials"),
        pretend.call("GitHub: environment doesn't look like a GH action; giving up"),
    ]


def test_detect_github_bad_permissions(monkeypatch):
    monkeypatch.setenv("GITHUB_ACTIONS", "true")
    monkeypatch.delenv("ACTIONS_ID_TOKEN_REQUEST_TOKEN", raising=False)
    monkeypatch.delenv("ACTIONS_ID_TOKEN_REQUEST_URL", raising=False)

    logger = pretend.stub(debug=pretend.call_recorder(lambda s: None))
    monkeypatch.setattr(ambient, "logger", logger)

    with pytest.raises(
        ambient.AmbientCredentialError,
        match="GitHub: missing or insufficient OIDC token permissions?",
    ):
        ambient.detect_github()
    assert logger.debug.calls == [
        pretend.call("GitHub: looking for OIDC credentials"),
    ]


def test_detect_github_request_fails(monkeypatch):
    monkeypatch.setenv("GITHUB_ACTIONS", "true")
    monkeypatch.setenv("ACTIONS_ID_TOKEN_REQUEST_TOKEN", "faketoken")
    monkeypatch.setenv("ACTIONS_ID_TOKEN_REQUEST_URL", "fakeurl")

    resp = pretend.stub(raise_for_status=pretend.raiser(HTTPError), status_code=999)
    requests = pretend.stub(
        get=pretend.call_recorder(lambda url, **kw: resp), HTTPError=HTTPError
    )
    monkeypatch.setattr(ambient, "requests", requests)

    with pytest.raises(
        ambient.AmbientCredentialError,
        match=r"GitHub: OIDC token request failed \(code=999\)",
    ):
        ambient.detect_github()
    assert requests.get.calls == [
        pretend.call(
            "fakeurl",
            params={"audience": "sigstore"},
            headers={"Authorization": "bearer faketoken"},
        )
    ]


def test_detect_github_bad_payload(monkeypatch):
    monkeypatch.setenv("GITHUB_ACTIONS", "true")
    monkeypatch.setenv("ACTIONS_ID_TOKEN_REQUEST_TOKEN", "faketoken")
    monkeypatch.setenv("ACTIONS_ID_TOKEN_REQUEST_URL", "fakeurl")

    resp = pretend.stub(
        raise_for_status=lambda: None, json=pretend.call_recorder(lambda: {})
    )
    requests = pretend.stub(get=pretend.call_recorder(lambda url, **kw: resp))
    monkeypatch.setattr(ambient, "requests", requests)

    with pytest.raises(
        ambient.AmbientCredentialError,
        match="GitHub: malformed or incomplete JSON",
    ):
        ambient.detect_github()
    assert requests.get.calls == [
        pretend.call(
            "fakeurl",
            params={"audience": "sigstore"},
            headers={"Authorization": "bearer faketoken"},
        )
    ]
    assert resp.json.calls == [pretend.call()]


def test_detect_github(monkeypatch):
    monkeypatch.setenv("GITHUB_ACTIONS", "true")
    monkeypatch.setenv("ACTIONS_ID_TOKEN_REQUEST_TOKEN", "faketoken")
    monkeypatch.setenv("ACTIONS_ID_TOKEN_REQUEST_URL", "fakeurl")

    resp = pretend.stub(
        raise_for_status=lambda: None,
        json=pretend.call_recorder(lambda: {"value": "fakejwt"}),
    )
    requests = pretend.stub(get=pretend.call_recorder(lambda url, **kw: resp))
    monkeypatch.setattr(ambient, "requests", requests)

    assert ambient.detect_github() == "fakejwt"
    assert requests.get.calls == [
        pretend.call(
            "fakeurl",
            params={"audience": "sigstore"},
            headers={"Authorization": "bearer faketoken"},
        )
    ]
    assert resp.json.calls == [pretend.call()]



#!/usr/bin/env python3
# Copyright 2022 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib

from setuptools import find_packages, setup

version = importlib.import_module("sigstore._version")

with open("./README.md") as f:
    long_description = f.read()

setup(
    name="sigstore",
    version=version.__version__,
    license="Apache-2.0",
    author="Sigstore Authors",
    author_email="sigstore-dev@googlegroups.com",
    description="A tool for signing Python package distributions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sigstore/sigstore-python",
    packages=find_packages(),
    package_data={"sigstore": ["_store/*"]},
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "sigstore = sigstore._cli:main",
        ]
    },
    platforms="any",
    python_requires=">=3.7",
    install_requires=[
        "click>=8",
        "cryptography",
        "pem",
        "pydantic",
        "pyjwt",
        "pyOpenSSL",
        "requests",
        "securesystemslib",
    ],
    extras_require={
        "dev": [
            "build",
            "bump",
            "flake8",
            "black",
            "isort",
            "pytest",
            "pytest-cov",
            "pretend",
            "coverage[toml]",
            "interrogate",
            "pdoc3",
            "mypy",
            "types-cryptography",
            "types-requests",
            "types-pyOpenSSL",
            "types-pyjwt",
        ]
    },
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Security",
        "Topic :: Security :: Cryptography",
    ],
)



# Copyright 2022 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from importlib import resources

import click

from sigstore._internal.oidc.ambient import detect_credential
from sigstore._internal.oidc.oauth import get_identity_token
from sigstore._sign import sign
from sigstore._verify import verify

logger = logging.getLogger(__name__)


@click.group()
def main():
    pass


@main.command("sign")
@click.option("identity_token", "--identity-token", type=click.STRING)
@click.option(
    "ctfe_pem",
    "--ctfe",
    type=click.File("rb"),
    default=resources.open_binary("sigstore._store", "ctfe.pub"),
)
@click.argument(
    "files", metavar="FILE [FILE ...]", type=click.File("rb"), nargs=-1, required=True
)
def _sign(files, identity_token, ctfe_pem):
    # The order of precedence is as follows:
    #
    # 1) Explicitly supplied identity token
    # 2) Ambient credential detected in the environment
    # 3) Interactive OAuth flow
    if not identity_token:
        identity_token = detect_credential()
    if not identity_token:
        identity_token = get_identity_token()

    ctfe_pem = ctfe_pem.read()
    for file in files:
        result = sign(
            file=file,
            identity_token=identity_token,
            ctfe_pem=ctfe_pem,
        )

        click.echo("Using ephemeral certificate:")
        click.echo(result.cert_pem)
        click.echo(
            f"Transparency log entry created at index: {result.log_entry.log_index}"
        )
        click.echo(f"Signature: {result.b64_signature}")


@main.command("verify")
@click.option("certificate_path", "--cert", type=click.File("rb"), required=True)
@click.option("signature_path", "--signature", type=click.File("rb"), required=True)
@click.option("cert_email", "--cert-email", type=str)
@click.argument(
    "files", metavar="FILE [FILE ...]", type=click.File("rb"), nargs=-1, required=True
)
def _verify(files, certificate_path, signature_path, cert_email):
    # Load the signing certificate
    logger.debug(f"Using certificate from: {certificate_path.name}")
    certificate = certificate_path.read()

    # Load the signature
    logger.debug(f"Using signature from: {signature_path.name}")
    signature = signature_path.read()

    for file in files:
        click.echo(
            verify(
                file=file,
                certificate=certificate,
                signature=signature,
                cert_email=cert_email,
                output=click.echo,
            )
        )
